module CAUL_CutTest

  ####################################
  # Author:        Anders Lyhagen
  ####################################

  @EPSILON = 0.00001
  @E_VE_POS = Geom::Vector3d.new(@EPSILON, @EPSILON, @EPSILON)
  @E_VE_NEG = Geom::Vector3d.new(-@EPSILON, -@EPSILON, -@EPSILON)
  @id_count #TODO: 64 bit int representation should be enforced
  @X_DIV = @Y_DIV = @Z_DIV = 10.0
  
  #Extract faces from g0 and g1 in the group bounding box intersection
  def self.getBBFaces(g0, g1)
    bb_inter = g0.bounds.intersect(g1.bounds)
    bb_min = bb_inter.min.offset(@E_VE_NEG)
    bb_max = bb_inter.max.offset(@E_VE_POS)
    g0bb = getBBArray(g0)
    g1bb = getBBArray(g1)   
    g0bb_int = g0bb.select { |b| bbIntersect?(bb_min, bb_max, b[0], b[1]) }
    g1bb_int = g1bb.select { |b| bbIntersect?(bb_min, bb_max, b[0], b[1]) }    
    return bb_inter, g0bb_int, g1bb_int
  end
  
  def self.getBBArray(g)
    tr = g.transformation
    return g.entities.grep(Sketchup::Face).map { |f|
      bb = Geom::BoundingBox.new
      bb.add(f.outer_loop.vertices.map { |v| v.position.transform tr })
      @id_count += 1
      [bb.min.offset(@E_VE_NEG), bb.max.offset(@E_VE_POS), f, @id_count] }
  end
   
  def self.bbIntersect?(a_min, a_max, b_min, b_max)
    return false if b_min.x > a_max.x || b_max.x < a_min.x
    return false if b_min.y > a_max.y || b_max.y < a_min.y
    return false if b_min.z > a_max.z || b_max.z < a_min.z
    return true
  end
  
  def self.extractFacePairs(bb, g0bb, g1bb)
    #compute the bounds for the space hash
    dx_inv = 1.0 / ((bb.max.x - bb.min.x) / @X_DIV)
    dy_inv = 1.0 / ((bb.max.y - bb.min.y) / @Y_DIV)
    dz_inv = 1.0 / ((bb.max.z - bb.min.z) / @Z_DIV)
    
    #create the space hash
    sh = []
    (0..@X_DIV - 1).each { |i| sh << []
      (0..@Y_DIV - 1).each { |j| sh[i] << []
        (0..@Z_DIV - 1).each { |k| sh[i][j] << []
    }}}
    
    #add g00b to space hash
    g0bb.each { |a| 
      inds = getBounds(a, bb, dx_inv, dy_inv, dz_inv)
      (inds[0]..inds[1]).each { |i|  
        (inds[2]..inds[3]).each { |j| 
          (inds[4]..inds[5]).each { |k|
            sh[i][j][k] << a
    }}}}
    
    res_h = {}
    res_a = []
    
    #find face pairs whose bounding boxes intersect
    g1bb.each { |b|
      b_id = (b[3] << 16) #TODO..make this <<32 ie. 64 bit
      inds = getBounds(b, bb, dx_inv, dy_inv, dz_inv)
      (inds[0]..inds[1]).each { |i|  
        (inds[2]..inds[3]).each { |j| 
          (inds[4]..inds[5]).each { |k|
            sh[i][j][k].each { |a| 
              next if res_h.has_key?(b_id + a[3])
              res_h[b_id + a[3]] = nil
              res_a << [a[2], b[2]] if bbIntersect?(a[0], a[1], b[0], b[1])
            }
      }}}
    }
    return res_a
  end
  
  def self.getBounds(b, bb, dx_inv, dy_inv, dz_inv)
    return [ [((b[0].x - bb.min.x) * dx_inv).floor, 0].max,
             [((b[1].x - bb.min.x) * dx_inv).floor, @X_DIV - 1].min,
             [((b[0].y - bb.min.y) * dy_inv).floor, 0].max,
             [((b[1].y - bb.min.y) * dy_inv).floor, @Y_DIV - 1].min,
             [((b[0].z - bb.min.z) * dz_inv).floor, 0].max,
             [((b[1].z - bb.min.z) * dz_inv).floor, @Z_DIV - 1].min ]
  end
  
  def self.intersectFacePairs(ent, g0tr, g1tr, fp)
    ng0 = ent.add_group
    ng1 = ent.add_group
    res_g = ent.add_group
    idt = Geom::Transformation.new
    
    fp.each { |fs|
      f0 = ng0.entities.add_face(fs[0].outer_loop.vertices.map { |v| v.position.transform g0tr })
      f1 = ng1.entities.add_face(fs[1].outer_loop.vertices.map { |v| v.position.transform g1tr })
      ng0.entities.intersect_with(false, idt, res_g ,idt , true, ng1)
      ng1.entities.intersect_with(false, idt, res_g ,idt , true, ng0) if coplanar?(f0, f1)
      ng0.entities.clear!
      ng1.entities.clear!
    }
    ng0.erase!
    ng1.erase!
    return res_g
  end
  
  def self.coplanar?(f0, f1)
    return false if f0.normal.dot(f1.normal).abs < 0.999
    vs = (f0.vertices + f1.vertices)
    pl = Geom::fit_plane_to_points vs  
    return vs.all? { |v| v.position.on_plane? pl }
  end
  
  def self.main
    mod = Sketchup.active_model
    ent = mod.entities
    sel = mod.selection
    
    gs = sel.grep(Sketchup::Group)
    return if gs.length != 2
    
    mod.start_operation("CutTest", true) 
    @id_count = 0
    bb, g0bb, g1bb = getBBFaces(gs[0], gs[1])
    fp = extractFacePairs(bb, g0bb, g1bb)
    g = intersectFacePairs(ent, gs[0].transformation, gs[1].transformation, fp)
    puts 'DONE'
    mod.commit_operation
  end
  
  unless file_loaded?( __FILE__ )
    menu = UI.menu( 'Plugins' )
    menu.add_item( 'CutTest' ) { main }
  end

  file_loaded( __FILE__ )
  
  main
end
